@@ -1,3 +1,24 @@
+/*
+ * Copyright 1999-2001 Red Hat, Inc.
+ * 
+ * All Rights Reserved.
+ * 
+ * The above copyright notice and this permission notice shall be included in
+ * all copies or substantial portions of the Software.
+ * 
+ * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
+ * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
+ * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
+ * OPEN GROUP BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
+ * AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
+ * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
+ * 
+ * Except as contained in this notice, the name of Red Hat shall not be
+ * used in advertising or otherwise to promote the sale, use or other dealings
+ * in this Software without prior written authorization from Red Hat.
+ *
+ */
+
 #include <arpa/inet.h>
 #include <ctype.h>
 #include <errno.h>
@@ -41,6 +62,7 @@ typedef short bp_int16;
 #define BOOTP_OPTION_BOOTFILE		13
 #define BOOTP_OPTION_DOMAIN		15
 #define BOOTP_OPTION_BROADCAST		28
+#define BOOTP_OPTION_NISDOMAIN		40
 
 #define DHCP_OPTION_LOGSRVS		7
 #define DHCP_OPTION_LPRSRVS		9
@@ -55,6 +77,7 @@ typedef short bp_int16;
 #define DHCP_OPTION_OPTIONREQ		55
 #define DHCP_OPTION_MAXSIZE		57
 #define DHCP_OPTION_T1			58
+#define DHCP_OPTION_CLIENT_IDENTIFIER	61
 
 #define BOOTP_CLIENT_PORT	68
 #define BOOTP_SERVER_PORT	67
@@ -83,7 +106,7 @@ struct bootpRequest {
     char hw;
     char hwlength;
     char hopcount;
-    bp_int32 id;
+    bp_int32 xid;
     bp_int16 secs;
     bp_int16 flags;
     bp_int32 ciaddr, yiaddr, server_ip, bootp_gw_ip;
@@ -109,16 +132,21 @@ static char * handleTransaction(int s, struct pumpOverrideInfo * override,
 				struct bootpRequest * breq,
 			        struct bootpRequest * bresp, 
 			        struct sockaddr_in * serverAddr,
-			        struct sockaddr_in * respondant,
-				int useBootPacket, int dhcpResponseType);
+				struct sockaddr_in * respondant,
+				int useBootpPacket, 
+				time_t startTime, int dhcpResponseType);
 static int dhcpMessageType(struct bootpRequest * response);
-static int newKernel(void);
+static int oldKernel(void);
 static char * getInterfaceInfo(struct pumpNetIntf * intf, int s);
 static char * perrorstr(char * msg);
+static void addClientIdentifier(int flags, struct bootpRequest * req);
+static void buildRequest(struct bootpRequest * req, int flags, int type,
+		         char * reqHostname, int lease);
+static void updateSecCount(struct bootpRequest * breq, time_t startTime);
 
 static const char vendCookie[] = { 99, 130, 83, 99, 255 };
 
-static int newKernel(void) {
+static int oldKernel(void) {
     struct utsname ubuf;
     int major1, major2;
 
@@ -126,17 +154,18 @@ static int newKernel(void) {
     if (!strcasecmp(ubuf.sysname, "linux")) {
 	if (sscanf(ubuf.release, "%d.%d", &major1, &major2) != 2 ||
 		(major1 < 2) || (major1 == 2 && major2 == 0)) {
-	    return 0;
+	    return 1;
 	}
     }
 
-    return 1;
+    return 0;
 }
 
 static char * getInterfaceInfo(struct pumpNetIntf * intf, int s) {
     struct ifreq req;
     struct sockaddr_in * addrp;
 
+    memset(&req,0,sizeof(req));
     strcpy(req.ifr_name, intf->device);
     if (ioctl(s, SIOCGIFBRDADDR, &req))
 	return perrorstr("SIOCGIFBRDADDR");
@@ -178,6 +207,8 @@ char * pumpDisableInterface(char * device) {
     int s;
 
     s = socket(AF_INET, SOCK_DGRAM, 0);
+	
+    memset(&req,0,sizeof(req));
 
     strcpy(req.ifr_name, device);
     if (ioctl(s, SIOCGIFFLAGS, &req)) {
@@ -206,7 +237,9 @@ char * pumpSetupInterface(struct pumpNetIntf * intf) {
     if ((rc = pumpDisableInterface(intf->device))) return rc;
 
     s = socket(AF_INET, SOCK_DGRAM, 0);
-
+	
+    memset(&req,0,sizeof(req));
+    memset(&route,0,sizeof(route));
     /* we have to have basic information to get this far */
     addrp = (struct sockaddr_in *) &req.ifr_addr;
     addrp->sin_family = AF_INET;
@@ -224,17 +257,22 @@ char * pumpSetupInterface(struct pumpNetIntf * intf) {
     if (ioctl(s, SIOCSIFBRDADDR, &req))
 	return perrorstr("SIOCSIFBRDADDR");
 
-    /* bring up the device, and specifically allow broadcasts through it */
-    req.ifr_flags = IFF_UP | IFF_RUNNING | IFF_BROADCAST;
+    /* Bring up the device, and specifically allow broadcasts through it.
+       Don't mess with flags we don't understand though. */
+    if (ioctl(s, SIOCGIFFLAGS, &req))
+	return perrorstr("SIOCGIFFLAGS");
+
+    req.ifr_flags |= IFF_UP | IFF_RUNNING | IFF_BROADCAST;
+
     if (ioctl(s, SIOCSIFFLAGS, &req))
 	return perrorstr("SIOCSIFFLAGS");
 
-    /* add a route for this network */
-    route.rt_dev = intf->device;
-    route.rt_flags = RTF_UP;
-    route.rt_metric = 0;
+    if (!strcmp(intf->device, "lo") || oldKernel()) {
+	/* add a route for this network */
+	route.rt_dev = intf->device;
+	route.rt_flags = RTF_UP;
+	route.rt_metric = 0;
 
-    if (!newKernel()) {
 	addrp->sin_family = AF_INET;
 	addrp->sin_port = 0;
 	addrp->sin_addr = intf->network;
@@ -257,7 +295,9 @@ int pumpSetupDefaultGateway(struct in_addr * gw) {
     int s;
 
     s = socket(AF_INET, SOCK_DGRAM, 0);
-
+	
+    memset(&addr,0,sizeof(addr));
+    memset(&route,0,sizeof(route));
     addr.sin_family = AF_INET;
     addr.sin_port = 0;
     addr.sin_addr.s_addr = INADDR_ANY;
@@ -271,6 +311,7 @@ int pumpSetupDefaultGateway(struct in_addr * gw) {
     route.rt_dev = NULL;
 
     if (ioctl(s, SIOCADDRT, &route)) {
+	syslog(LOG_ERR, "failed to set default route: %s", strerror(errno));
 	return -1;
     }
 
@@ -282,6 +323,8 @@ char * pumpPrepareInterface(struct pumpNetIntf * intf, int s) {
     struct ifreq req;
     struct rtentry route;
 
+    memset(&req,0,sizeof(req));
+
     addrp = (struct sockaddr_in *) &req.ifr_addr;
 
     strcpy(req.ifr_name, intf->device);
@@ -295,7 +338,7 @@ char * pumpPrepareInterface(struct pumpNetIntf * intf, int s) {
     if (ioctl(s, SIOCSIFADDR, &req))
 	return perrorstr("SIOCSIFADDR");
 
-    if (!newKernel()) {
+    if (oldKernel()) {
 	if (ioctl(s, SIOCSIFNETMASK, &req))
 	    return perrorstr("SIOCSIFNETMASK");
 
@@ -305,7 +348,9 @@ char * pumpPrepareInterface(struct pumpNetIntf * intf, int s) {
 	    return perrorstr("SIOCSIFBRDADDR");
     }
 
-    req.ifr_flags = IFF_UP | IFF_BROADCAST | IFF_RUNNING;
+    if (ioctl(s, SIOCGIFFLAGS, &req))
+	return perrorstr("SIOCGIFFLAGS");
+    req.ifr_flags |= IFF_UP | IFF_BROADCAST | IFF_RUNNING;
     if (ioctl(s, SIOCSIFFLAGS, &req))
 	return perrorstr("SIOCSIFFLAGS");
 
@@ -347,6 +392,7 @@ static int dhcpMessageType(struct bootpRequest * response) {
 	    return *chptr;
 
 	chptr += length;
+	if ( ((void *)chptr - (void *)response->vendor) >= DHCP_VENDOR_LENGTH) return -1;
     }
 
     return -1;
@@ -399,10 +445,10 @@ static void parseReply(struct bootpRequest * breq, struct pumpNetIntf * intf) {
     if (strlen(breq->bootfile)) {
 	intf->bootFile = strdup(breq->bootfile);
 	intf->set |= PUMP_INTFINFO_HAS_BOOTFILE;
+	syslog (LOG_DEBUG, "intf: bootFile: %s", intf->bootFile);
     } else {
 	intf->set &= ~PUMP_INTFINFO_HAS_BOOTFILE;
     }
-    syslog (LOG_DEBUG, "intf: bootFile: %s", intf->bootFile);
 
     memcpy(&intf->ip, &breq->yiaddr, 4);
     intf->set |= PUMP_INTFINFO_HAS_IP;
@@ -439,6 +485,15 @@ static void parseReply(struct bootpRequest * breq, struct pumpNetIntf * intf) {
 		syslog (LOG_DEBUG, "intf: netmask: %s", inet_ntoa (intf->netmask));
 		break;
 
+	    case BOOTP_OPTION_NISDOMAIN:
+		if ((intf->nisDomain = malloc(length + 1))) {
+		    memcpy(intf->nisDomain, chptr, length);
+		    intf->nisDomain[length] = '\0';
+		    intf->set |= PUMP_NETINFO_HAS_NISDOMAIN;
+		    syslog (LOG_DEBUG, "intf: nisDomain: %s", intf->nisDomain);
+		}
+		break;
+
 	    case BOOTP_OPTION_DOMAIN:
 		if ((intf->domain = malloc(length + 1))) {
 		    memcpy(intf->domain, chptr, length);
@@ -559,6 +614,7 @@ static char * prepareRequest(struct bootpRequest * breq,
     int i;
 
     memset(breq, 0, sizeof(*breq));
+    memset(&req, 0, sizeof(req));
 
     breq->opcode = BOOTP_OPCODE_REQUEST;
 
@@ -566,24 +622,28 @@ static char * prepareRequest(struct bootpRequest * breq,
     if (ioctl(sock, SIOCGIFHWADDR, &req))
 	return perrorstr("SIOCSIFHWADDR");
 
-    breq->hw = 1; 		/* ethernet */
+    breq->hw = req.ifr_hwaddr.sa_family;
     breq->hwlength = IFHWADDRLEN;	
     memcpy(breq->hwaddr, req.ifr_hwaddr.sa_data, IFHWADDRLEN);
 
     /* we should use something random here, but I don't want to start using
        stuff from the math library */
-    breq->id = time(NULL);
+    breq->xid = pumpUptime();
     for (i = 0; i < IFHWADDRLEN; i++)
-	breq->id ^= breq->hwaddr[i] << (8 * (i % 4));
+	breq->xid ^= breq->hwaddr[i] << (8 * (i % 4));
 
     breq->hopcount = 0;
-    breq->secs = time(NULL) - startTime;
+    breq->secs = 0;
 
     initVendorCodes(breq);
 
     return NULL;
 }
 
+static void updateSecCount(struct bootpRequest * breq, time_t startTime) {
+    breq->secs = pumpUptime() - startTime;
+}
+
 static unsigned int verifyChecksum(void * buf, int length, void * buf2,
 				   int length2) {
     unsigned int csum;
@@ -617,13 +677,15 @@ void debugbootpRequest(char *name, struct bootpRequest *breq)  {
     unsigned char *vndptr;
     unsigned char option, length;
     
+    memset(&address,0,sizeof(address));
+	
     syslog (LOG_DEBUG, "%s: opcode: %i", name, breq->opcode);
     syslog (LOG_DEBUG, "%s: hw: %i", name, breq->hw);
     syslog (LOG_DEBUG, "%s: hwlength: %i", name, breq->hwlength);
     syslog (LOG_DEBUG, "%s: hopcount: %i", name, breq->hopcount);
-    syslog (LOG_DEBUG, "%s: id: 0x%8x", name, breq->id);
+    syslog (LOG_DEBUG, "%s: xid: 0x%08x", name, breq->xid);
     syslog (LOG_DEBUG, "%s: secs: %i", name, breq->secs);
-    syslog (LOG_DEBUG, "%s: flags: 0x%4x", name, breq->flags);
+    syslog (LOG_DEBUG, "%s: flags: 0x%04x", name, breq->flags);
     
     address.s_addr = breq->ciaddr;
     syslog (LOG_DEBUG, "%s: ciaddr: %s", name, inet_ntoa (address));
@@ -642,7 +704,7 @@ void debugbootpRequest(char *name, struct bootpRequest *breq)  {
     syslog (LOG_DEBUG, "%s: bootfile: %s", name, breq->bootfile);
     
     vndptr = breq->vendor;
-    sprintf (vendor, "0x%2x 0x%2x 0x%2x 0x%2x", *vndptr++, *vndptr++, *vndptr++, *vndptr++);
+    sprintf (vendor, "0x%02x 0x%02x 0x%02x 0x%02x", *vndptr++, *vndptr++, *vndptr++, *vndptr++);
     syslog (LOG_DEBUG, "%s: vendor: %s", name, vendor);
     
     
@@ -651,13 +713,13 @@ void debugbootpRequest(char *name, struct bootpRequest *breq)  {
 	option = *vndptr++;
 	if (option == 0xFF)
 	  {
-	    sprintf (vendor, "0x%2x", option);
+	    sprintf (vendor, "0x%02x", option);
 	    vndptr = breq->vendor + DHCP_VENDOR_LENGTH;
 	  }
 	else if (option == 0x00)
 	  {
 	    for (i = 1; *vndptr == 0x00; i++, vndptr++);
-	    sprintf (vendor, "0x%2x x %i", option, i);
+	    sprintf (vendor, "0x%02x x %i", option, i);
 	  }
 	else
 	  {
@@ -670,7 +732,7 @@ void debugbootpRequest(char *name, struct bootpRequest *breq)  {
 		    syslog (LOG_DEBUG, "%s: vendor: %s", name, vendor);
 		    strcpy (vendor, "++++++");
 		  }
-		snprintf (vendor2, 27, "%s 0x%2x", vendor, *vndptr++);
+		snprintf (vendor2, 27, "%s 0x%02x", vendor, *vndptr++);
 		strcpy (vendor, vendor2);
 		
 	      }
@@ -688,7 +750,8 @@ static char * handleTransaction(int s, struct pumpOverrideInfo * override,
 			        struct bootpRequest * bresp, 
 			        struct sockaddr_in * serverAddr,
 				struct sockaddr_in * respondant,
-				int useBootpPacket, int dhcpResponseType) {
+				int useBootpPacket, 
+				time_t startTime, int dhcpResponseType) {
     struct timeval tv;
     fd_set readfs;
     int i, j;
@@ -701,11 +764,12 @@ static char * handleTransaction(int s, struct pumpOverrideInfo * override,
     int resend = 1;
     struct ethhdr;
     char ethPacket[ETH_FRAME_LEN];
-    struct iphdr * ipHdr;
-    struct udphdr * udpHdr;
+    struct iphdr * ipHdr = NULL;
+    struct udphdr * udpHdr = NULL;
     struct psuedohUdpHeader pHdr;
-    time_t start = time(NULL);
+    time_t start = pumpUptime();
     
+    memset(&pHdr,0,sizeof(pHdr));
     debugbootpRequest("breq", breq);
 
     if (!override) {
@@ -726,6 +790,9 @@ static char * handleTransaction(int s, struct pumpOverrideInfo * override,
 	    i -= (DHCP_VENDOR_LENGTH - BOOTP_VENDOR_LENGTH);
 
 	if (resend) {
+	    if (startTime != -1)
+		updateSecCount(breq, startTime);
+
 	    if (sendto(s, breq, i, 0, (struct sockaddr *) serverAddr, 
 		       sizeof(*serverAddr)) != i) {
 		close(sin);
@@ -735,12 +802,12 @@ static char * handleTransaction(int s, struct pumpOverrideInfo * override,
 	    tries--;
 	    nextTimeout *= 2;
 
-	    switch (time(NULL) & 4) {
+	    switch (pumpUptime() & 4) {
 		case 0:	if (nextTimeout >= 2) nextTimeout--; break;
 		case 1:	nextTimeout++; break;
 	    }
 
-	    timeoutTime = time(NULL) + nextTimeout;
+	    timeoutTime = pumpUptime() + nextTimeout;
 	    i = override->timeout + start;
 	    if (timeoutTime > i) timeoutTime = i;
 
@@ -753,8 +820,8 @@ static char * handleTransaction(int s, struct pumpOverrideInfo * override,
 	}
 
 	tv.tv_usec = 0;
- 	tv.tv_sec = timeoutTime - time(NULL);
-	if (timeoutTime < time(NULL)) {
+ 	tv.tv_sec = timeoutTime - pumpUptime();
+	if (timeoutTime < pumpUptime()) {
 	    tries = 0;
 	    continue;
 	}
@@ -779,6 +846,10 @@ static char * handleTransaction(int s, struct pumpOverrideInfo * override,
 	    if (!verifyChecksum(NULL, 0, ipHdr, sizeof(*ipHdr)))
 		continue;
 
+	    /* XXX the broadcast must always be from 0.0.0.0, so we'll
+	       set that now. */
+	    ipHdr->saddr = 0;
+	    
 	    if (ntohs(ipHdr->tot_len) > j)
 		continue;
 	    j = ntohs(ipHdr->tot_len);
@@ -816,14 +887,39 @@ static char * handleTransaction(int s, struct pumpOverrideInfo * override,
 	  memcpy (bresp, (char *) udpHdr + sizeof (*udpHdr), j - sizeof (*ipHdr) - sizeof (*udpHdr));
 
 	    /* sanity checks */
-	    if (bresp->id != breq->id) continue;
-	    if (bresp->opcode != BOOTP_OPCODE_REPLY) continue;
-	    if (bresp->hwlength != breq->hwlength) continue;
-	    if (memcmp(bresp->hwaddr, breq->hwaddr, bresp->hwlength)) continue;
-	    i = dhcpMessageType(bresp);
-	    if (!(i == -1 && useBootpPacket) && (i != dhcpResponseType))
+	    if (bresp->xid != breq->xid) {
+		syslog(LOG_DEBUG, "reject: xid: 0x%08x <--> 0x%08x",
+			breq->xid, bresp->xid);
 		continue;
-	    if (memcmp(bresp->vendor, vendCookie, 4)) continue;
+	    }
+	    if (bresp->opcode != BOOTP_OPCODE_REPLY) {
+		syslog(LOG_DEBUG, "reject: opcode: %i <--> %i",
+			BOOTP_OPCODE_REPLY, bresp->opcode);
+		continue;
+	    }
+	    if (bresp->hwlength != breq->hwlength) {
+		syslog(LOG_DEBUG, "reject: hwlength: %i <--> %i",
+			breq->hwlength, bresp->hwlength);
+		continue;
+	    }
+	    if (memcmp(bresp->hwaddr, breq->hwaddr, bresp->hwlength)) {
+		syslog(LOG_DEBUG, "reject: hwaddr: %s <--> %s",
+			breq->hwaddr, bresp->hwaddr);
+		continue;
+	    }
+	    i = dhcpMessageType(bresp);
+	    if (!(i == -1 && useBootpPacket) && (i != dhcpResponseType)) {
+		syslog(LOG_DEBUG, "reject: msgtyp: %i", i);
+		continue;
+	    }
+	    if (memcmp(bresp->vendor, vendCookie, 4)) {
+		syslog(LOG_DEBUG, "reject: vendor: 0x%02x 0x%02x 0x%02x 0x%02x"
+				  " <--> 0x%02x 0x%02x 0x%02x 0x%02x",
+		       vendCookie[0], vendCookie[1], vendCookie[2], 
+		       vendCookie[3], bresp->vendor[0], bresp->vendor[1],
+		       bresp->vendor[2], bresp->vendor[3]);
+		continue;
+	    }
 	    /* if (respondant) *respondant = tmpAddress; */
 	    gotit = 1;
 
@@ -868,7 +964,7 @@ static void addVendorCode(struct bootpRequest * breq, unsigned char option,
 }
 
 static int getVendorCode(struct bootpRequest * bresp, unsigned char option,
-			  void * data) {
+			  void * data, size_t maxsize) {
     unsigned char * chptr;
     unsigned int length, theOption;
 
@@ -879,17 +975,20 @@ static int getVendorCode(struct bootpRequest * bresp, unsigned char option,
 	if (!theOption) continue;
 	length = *chptr++;
 	chptr += length;
+	if ( ((void *)chptr - (void *)bresp->vendor) >= DHCP_VENDOR_LENGTH) return 1;
     }
 
     if (*chptr++ == 0xff) return 1;
 
     length = *chptr++;
+    if(length > maxsize)
+        length = maxsize;
     memcpy(data, chptr, length);
 
     return 0;
 }
 
-static int createSocket(void) {
+static int createSocket(const char * device) {
     struct sockaddr_in clientAddr;
     int s;
     int true = 1;
@@ -903,6 +1002,10 @@ static int createSocket(void) {
 	return -1;
     }
 
+    if (setsockopt(s, SOL_SOCKET, SO_BINDTODEVICE, device, strlen(device)+1)) {
+	syslog(LOG_ERR, "SO_BINDTODEVICE %s (%d) failed: %s", device, strlen(device), strerror(errno));
+    }
+
     memset(&clientAddr.sin_addr, 0, sizeof(&clientAddr.sin_addr));
     clientAddr.sin_family = AF_INET;
     clientAddr.sin_port = htons(BOOTP_CLIENT_PORT);	/* bootp client */
@@ -921,6 +1024,7 @@ int pumpDhcpRelease(struct pumpNetIntf * intf) {
     struct sockaddr_in serverAddr;
     char * chptr;
     int s;
+    char hostname[1024];
 
     if (!(intf->set & PUMP_INTFINFO_HAS_LEASE)) {
 	pumpDisableInterface(intf->device);
@@ -929,9 +1033,9 @@ int pumpDhcpRelease(struct pumpNetIntf * intf) {
 	return 0;
     }
 
-    if ((s = createSocket()) < 0) return 1;
+    if ((s = createSocket(intf->device)) < 0) return 1;
 
-    if ((chptr = prepareRequest(&breq, s, intf->device, time(NULL)))) {
+    if ((chptr = prepareRequest(&breq, s, intf->device, pumpUptime()))) {
 	close(s);
 	while (1) {
 	    pumpDisableInterface(intf->device);
@@ -943,23 +1047,37 @@ int pumpDhcpRelease(struct pumpNetIntf * intf) {
     addVendorCode(&breq, DHCP_OPTION_TYPE, 1, &messageType);
     memcpy(&breq.ciaddr, &intf->ip, sizeof(breq.ciaddr));
 
+    /* Dynamic DHCP implementations need the hostname here. */
+    if (intf->set & PUMP_NETINFO_HAS_HOSTNAME) {
+	addVendorCode(&breq, BOOTP_OPTION_HOSTNAME,
+		      strlen(intf->hostname) + 1, intf->hostname);
+    } else {
+	gethostname(hostname, sizeof(hostname));
+	if (strcmp(hostname, "localhost") && 
+	    strcmp(hostname, "localhost.localdomain")) {
+	    addVendorCode(&breq, BOOTP_OPTION_HOSTNAME, 
+			  strlen(hostname) + 1, hostname);
+	}
+    }
+
     serverAddr.sin_family = AF_INET;
     serverAddr.sin_port = htons(BOOTP_SERVER_PORT);	/* bootp server */
     serverAddr.sin_addr = intf->bootServer;
 
-    if (!handleTransaction(s, NULL, &breq, &bresp, &serverAddr, NULL, 0,
-			   NORESPONSE)) {
-	pumpDisableInterface(intf->device);
-	close(s);
-	return 0;
-    }
+    handleTransaction(s, NULL, &breq, &bresp, &serverAddr, NULL, 0, 
+		      -1, NORESPONSE);
 
     pumpDisableInterface(intf->device);
     close(s);
 
+    if (intf->set & PUMP_NETINFO_HAS_HOSTNAME)
+	free(intf->hostname);
+    if (intf->set & PUMP_NETINFO_HAS_DOMAIN)
+	free(intf->domain);
+
     syslog(LOG_INFO, "disabling interface %s", intf->device);
 
-    return 1;
+    return 0;
 }
     
 /* This is somewhat broken. We try only to renew the lease. If we fail,
@@ -973,10 +1091,11 @@ int pumpDhcpRenew(struct pumpNetIntf * intf) {
     int s;
     int i;
     char hostname[1024];
+    time_t startTime = pumpUptime();
 
-    s = createSocket();
+    s = createSocket(intf->device);
 
-    if ((chptr = prepareRequest(&breq, s, intf->device, time(NULL)))) {
+    if ((chptr = prepareRequest(&breq, s, intf->device, pumpUptime()))) {
 	close(s);
 	while (1);	/* problem */
     }
@@ -985,16 +1104,19 @@ int pumpDhcpRenew(struct pumpNetIntf * intf) {
     addVendorCode(&breq, DHCP_OPTION_TYPE, 1, &messageType);
     memcpy(&breq.ciaddr, &intf->ip, sizeof(breq.ciaddr));
 
+    addClientIdentifier(intf->flags, &breq);
+
     /* Dynamic DHCP implementations need the hostname here. */
     if (intf->set & PUMP_NETINFO_HAS_HOSTNAME) {
-	addVendorCode(&breq, BOOTP_OPTION_HOSTNAME, strlen(intf->hostname),
+	addVendorCode(&breq, BOOTP_OPTION_HOSTNAME,
+		      strlen(intf->hostname) + 1,
 		      intf->hostname);
     } else {
 	gethostname(hostname, sizeof(hostname));
 	if (strcmp(hostname, "localhost") && 
 	    strcmp(hostname, "localhost.localdomain")) {
 	    addVendorCode(&breq, BOOTP_OPTION_HOSTNAME, 
-			  strlen(hostname), hostname);
+			  strlen(hostname) + 1, hostname);
 	}
     }
 
@@ -1006,7 +1128,7 @@ int pumpDhcpRenew(struct pumpNetIntf * intf) {
     serverAddr.sin_addr = intf->bootServer;
 
     if (handleTransaction(s, NULL, &breq, &bresp, &serverAddr, NULL, 0,
-			   DHCP_TYPE_ACK)) {
+			   startTime, DHCP_TYPE_ACK)) {
 	close(s);
 	return 1;
     }
@@ -1024,35 +1146,105 @@ static void parseLease(struct bootpRequest * bresp, struct pumpNetIntf * intf) {
     time_t now;
 
     intf->set &= ~PUMP_INTFINFO_HAS_LEASE;
-    if (getVendorCode(bresp, DHCP_OPTION_LEASE, &lease)) 
+    if (getVendorCode(bresp, DHCP_OPTION_LEASE, &lease, sizeof(int))) 
 	return;
 
     lease = ntohl(lease);
 
     if (lease && lease != 0xffffffff) {
-	now = time(NULL);
+	now = pumpUptime();
 	intf->set |= PUMP_INTFINFO_HAS_LEASE;
 	intf->leaseExpiration = now + lease;
 	intf->renewAt = now + (7 * lease / 8);
     }
 }
 
+static void addClientIdentifier(int flags, struct bootpRequest * req) {
+    char ident[256];
+
+    /*
+     * Microsoft uses a client identifier field of the 802.3 address with a
+     * pre-byte of a "1".  In order to re-use the DHCP address that they set
+     * for this interface, we have to mimic their identifier.
+    */
+
+    if (!(flags & PUMP_FLAG_WINCLIENTID)) return;
+
+    ident[0] = 1;
+    memcpy(ident + 1, req->hwaddr, req->hwlength);
+    addVendorCode(req, DHCP_OPTION_CLIENT_IDENTIFIER, req->hwlength + 1, 
+		  ident);
+}
+
+static void buildRequest(struct bootpRequest * req, int flags, int type,
+		         char * reqHostname, int lease) {
+    unsigned char messageType = type;
+    short aShort;
+    int anInt;
+    int numOptions;
+    char optionsRequested[50];
+    static int aCount = 1;
+
+    addVendorCode(req, DHCP_OPTION_TYPE, 1, &messageType);
+
+    addClientIdentifier(flags, req);
+
+    aShort = ntohs(sizeof(struct bootpRequest));
+    addVendorCode(req, DHCP_OPTION_MAXSIZE, 2, &aShort);
+
+    numOptions = 0;
+    optionsRequested[numOptions++] = BOOTP_OPTION_NETMASK;
+    optionsRequested[numOptions++] = BOOTP_OPTION_GATEWAY;
+    optionsRequested[numOptions++] = BOOTP_OPTION_DNS;
+    optionsRequested[numOptions++] = BOOTP_OPTION_DOMAIN;
+    optionsRequested[numOptions++] = BOOTP_OPTION_BROADCAST;
+    optionsRequested[numOptions++] = BOOTP_OPTION_HOSTNAME;
+    optionsRequested[numOptions++] = DHCP_OPTION_LOGSRVS;
+    optionsRequested[numOptions++] = DHCP_OPTION_LPRSRVS;
+    optionsRequested[numOptions++] = DHCP_OPTION_NTPSRVS;
+    optionsRequested[numOptions++] = DHCP_OPTION_XFNTSRVS;
+    optionsRequested[numOptions++] = DHCP_OPTION_XDMSRVS;
+    addVendorCode(req, DHCP_OPTION_OPTIONREQ, numOptions, 
+		  optionsRequested);
+
+    req->xid += aCount;
+
+    if (!reqHostname) {
+	reqHostname = alloca(200);
+	gethostname(reqHostname, 200);
+	if (!strcmp(reqHostname, "localhost") ||
+	    !strcmp(reqHostname, "localhost.localdomain"))
+	    reqHostname = NULL;
+    }
+
+    if (reqHostname) {
+	addVendorCode(req, BOOTP_OPTION_HOSTNAME,
+		      strlen(reqHostname) + 1, reqHostname);
+    }
+
+    anInt = htonl(lease);
+    addVendorCode(req, DHCP_OPTION_LEASE, 4, &anInt);
+}
+
 char * pumpDhcpRun(char * device, int flags, int reqLease,
 		   char * reqHostname, struct pumpNetIntf * intf,
 		   struct pumpOverrideInfo * override) {
-    int s, i;
+    int s;
     struct sockaddr_in serverAddr;
     struct sockaddr_in clientAddr;
     struct sockaddr_in broadcastAddr;
     struct bootpRequest breq, bresp;
-    struct bootpRequest protoReq;
     unsigned char * chptr;
-    unsigned char messageType;
-    time_t startTime = time(NULL);
+    time_t startTime = pumpUptime();
     int true = 1;
-    char optionsRequested[50];
-    int numOptions;
-    short aShort;
+    char * saveDeviceName;
+    unsigned char messageType;
+
+    /* If device is the same as intf->device, don't let the memset()
+       blow away the device name */
+    saveDeviceName = alloca(strlen(device) + 1);
+    strcpy(saveDeviceName, device);
+    device = saveDeviceName;
 
     memset(intf, 0, sizeof(*intf));
     strcpy(intf->device, device);
@@ -1088,6 +1280,12 @@ char * pumpDhcpRun(char * device, int flags, int reqLease,
     messageType = DHCP_TYPE_DISCOVER;
     addVendorCode(&breq, DHCP_OPTION_TYPE, 1, &messageType);
 
+    if (reqHostname) {
+	syslog(LOG_DEBUG, "HOSTNAME: requesting %s\n", reqHostname);
+	addVendorCode(&breq, BOOTP_OPTION_HOSTNAME,
+		      strlen(reqHostname) + 1, reqHostname);
+    }
+
     memset(&clientAddr.sin_addr, 0, sizeof(&clientAddr.sin_addr));
     clientAddr.sin_family = AF_INET;
     clientAddr.sin_port = htons(BOOTP_CLIENT_PORT);	/* bootp client */
@@ -1098,6 +1296,7 @@ char * pumpDhcpRun(char * device, int flags, int reqLease,
 	return perrorstr("bind");
     }
 
+    memset(&serverAddr,0,sizeof(serverAddr));
     serverAddr.sin_family = AF_INET;
     serverAddr.sin_port = htons(BOOTP_SERVER_PORT);	/* bootp server */
 
@@ -1107,6 +1306,7 @@ char * pumpDhcpRun(char * device, int flags, int reqLease,
 	serverAddr.sin_addr = intf->bootServer;
 #endif 
 
+    memset(&broadcastAddr,0,sizeof(broadcastAddr));
     broadcastAddr.sin_family = AF_INET;
     broadcastAddr.sin_port = htons(BOOTP_SERVER_PORT);
 
@@ -1122,7 +1322,7 @@ char * pumpDhcpRun(char * device, int flags, int reqLease,
     syslog (LOG_DEBUG, "PUMP: sending discover\n");
 
     if ((chptr = handleTransaction(s, override, &breq, &bresp, &broadcastAddr,
-				   NULL, 1, DHCP_TYPE_OFFER))) {
+				   NULL, 1, startTime, DHCP_TYPE_OFFER))) {
 	close(s);
 	pumpDisableInterface(intf->device);
 	return chptr;
@@ -1136,53 +1336,15 @@ char * pumpDhcpRun(char * device, int flags, int reqLease,
 	syslog (LOG_DEBUG, "got dhcp offer\n");
 
 	initVendorCodes(&breq);
-
-	aShort = ntohs(sizeof(struct bootpRequest));
-	addVendorCode(&breq, DHCP_OPTION_MAXSIZE, 2, &aShort);
-
-	numOptions = 0;
-	optionsRequested[numOptions++] = BOOTP_OPTION_NETMASK;
-	optionsRequested[numOptions++] = BOOTP_OPTION_GATEWAY;
-	optionsRequested[numOptions++] = BOOTP_OPTION_DNS;
-	optionsRequested[numOptions++] = BOOTP_OPTION_DOMAIN;
-	optionsRequested[numOptions++] = BOOTP_OPTION_BROADCAST;
-	optionsRequested[numOptions++] = BOOTP_OPTION_HOSTNAME;
-	optionsRequested[numOptions++] = DHCP_OPTION_LOGSRVS;
-	optionsRequested[numOptions++] = DHCP_OPTION_LPRSRVS;
-	optionsRequested[numOptions++] = DHCP_OPTION_NTPSRVS;
-	optionsRequested[numOptions++] = DHCP_OPTION_XFNTSRVS;
-	optionsRequested[numOptions++] = DHCP_OPTION_XDMSRVS;
-	addVendorCode(&breq, DHCP_OPTION_OPTIONREQ, numOptions, 
-		      optionsRequested);
-
-	if (!reqHostname) {
-	    reqHostname = alloca(200);
-	    gethostname(reqHostname, 200);
-	    if (!strcmp(reqHostname, "localhost") ||
-		!strcmp(reqHostname, "localhost.localdomain"))
-	        reqHostname = NULL;
-	}
-
-	if (reqHostname) {
-	    syslog(LOG_DEBUG, "HOSTNAME: requesting %s\n", reqHostname);
-	    addVendorCode(&breq, BOOTP_OPTION_HOSTNAME, strlen(reqHostname), 
-			  reqHostname);
-	}
-
-	i = htonl(intf->reqLease);
-	addVendorCode(&breq, DHCP_OPTION_LEASE, 4, &i);
-
-	protoReq = breq;
+	buildRequest(&breq, flags, DHCP_TYPE_DISCOVER, reqHostname, 
+		     intf->reqLease);
 
 	syslog (LOG_DEBUG, "PUMP: sending second discover");
 
-	messageType = DHCP_TYPE_DISCOVER;
-	addVendorCode(&breq, DHCP_OPTION_TYPE, 1, &messageType);
-
-	/* Send another DHCP_REQUEST with the proper option list */
+	/* Send another DHCP_DISCOVER with the proper option list */
 	if ((chptr = handleTransaction(s, override, &breq, &bresp, 
-				       &broadcastAddr, NULL, 1, 
-				       DHCP_TYPE_OFFER))) {
+				       &broadcastAddr, NULL, 0, 
+				       startTime, DHCP_TYPE_OFFER))) {
 	    close(s);
 	    pumpDisableInterface(intf->device);
 	    return chptr;
@@ -1197,14 +1359,14 @@ char * pumpDhcpRun(char * device, int flags, int reqLease,
 
 	syslog (LOG_DEBUG, "PUMP: got an offer");
 
-	if (getVendorCode(&bresp, DHCP_OPTION_SERVER, &serverAddr.sin_addr)) {
+	if (getVendorCode(&bresp, DHCP_OPTION_SERVER, &serverAddr.sin_addr, sizeof(struct in_addr))) {
 	    syslog (LOG_DEBUG, "DHCPOFFER didn't include server address");
 	    intf->bootServer = broadcastAddr.sin_addr;
 	}
 
-	breq = protoReq;
-	messageType = DHCP_TYPE_REQUEST;
-	addVendorCode(&breq, DHCP_OPTION_TYPE, 1, &messageType);
+	initVendorCodes(&breq);
+	buildRequest(&breq, flags, DHCP_TYPE_REQUEST, reqHostname, 
+		     intf->reqLease);
 
 	addVendorCode(&breq, DHCP_OPTION_SERVER, 4, &serverAddr.sin_addr);
 	addVendorCode(&breq, DHCP_OPTION_REQADDR, 4, &bresp.yiaddr);
@@ -1213,7 +1375,7 @@ char * pumpDhcpRun(char * device, int flags, int reqLease,
 	   spec! */
 	if ((chptr = handleTransaction(s, override, &breq, &bresp, 
 				       &broadcastAddr, NULL, 0, 
-				       DHCP_TYPE_ACK))) {
+				       startTime, DHCP_TYPE_ACK))) {
 	    close(s);
 	    pumpDisableInterface(intf->device);
 	    return chptr;
@@ -1223,7 +1385,7 @@ char * pumpDhcpRun(char * device, int flags, int reqLease,
 
 	parseLease(&bresp, intf);
 
-	if (getVendorCode(&bresp, DHCP_OPTION_SERVER, &intf->bootServer)) {
+	if (getVendorCode(&bresp, DHCP_OPTION_SERVER, &intf->bootServer, sizeof(struct in_addr))) {
 	    syslog (LOG_DEBUG, "DHCPACK didn't include server address");
 	    intf->bootServer = broadcastAddr.sin_addr;
 	}
@@ -1237,6 +1399,9 @@ char * pumpDhcpRun(char * device, int flags, int reqLease,
     if (flags & PUMP_FLAG_FORCEHNLOOKUP)
 	intf->set &= ~(PUMP_NETINFO_HAS_DOMAIN | PUMP_NETINFO_HAS_HOSTNAME);
 
+    /* Save these for later */
+    intf->flags = flags & PUMP_FLAG_WINCLIENTID;
+
     return NULL;
 }
 
@@ -1244,5 +1409,39 @@ void pumpInitOverride(struct pumpOverrideInfo * override) {
     strcpy(override->intf.device, "MASTER");
     override->timeout = DEFAULT_TIMEOUT;
     override->numRetries = DEFAULT_NUM_RETRIES;
+    override->script = NULL;
+}
+
+/*
+* If time(2) is used, changing the date on a system can cause
+* pump to miss the lease renewal time.  The result is a system
+* that doesn't seem to want to talk until it is rebooted!  We
+* need elapsed time, NOT absolute (wall clock) time to measure
+* the expiration of the lease.  Unfortunately I can't see any
+* other interface to an unmolested elapsed time value other than
+* /proc/uptime (and this is how uptime(1), top(1), w(1) use it).
+* No matter, overhead is not an issue here, accuracy is.
+* duanev@io.com 9/99
+*/
+time_t pumpUptime() {
+    FILE * fd;
+    long secs;
+    static int first = 1;
+
+    if ((fd = fopen("/proc/uptime", "r")) == NULL) {
+	if (first) {
+	    syslog (LOG_INFO, "error opening /proc/uptime: %s", 
+		    strerror(errno));
+	    syslog (LOG_INFO, "warning: might miss lease renewal if date "
+		    "changes");
+	    first = 0;
+	}
+	return time(NULL);
+    }
+
+    fscanf(fd, "%ld", &secs);
+    fclose(fd);
+
+    return (time_t)secs;
 }
 
