@@ -1,3 +1,24 @@
+/*
+ * Copyright 1999-2001 Red Hat, Inc.
+ * 
+ * All Rights Reserved.
+ * 
+ * The above copyright notice and this permission notice shall be included in
+ * all copies or substantial portions of the Software.
+ * 
+ * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
+ * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
+ * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
+ * OPEN GROUP BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
+ * AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
+ * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
+ * 
+ * Except as contained in this notice, the name of Red Hat shall not be
+ * used in advertising or otherwise to promote the sale, use or other dealings
+ * in this Software without prior written authorization from Red Hat.
+ *
+ */
+
 #include <arpa/inet.h>
 #include <errno.h>
 #include <popt.h>
@@ -13,16 +34,18 @@
 #include <locale.h>
 
 #define _(String) gettext((String))
+#define N_(String) String
 
 #define LOADER_BACK 2
+#define LOADER_OK 0
 #define LOADER_ERROR -1;
 
 #include "net.h"
+#include "pump.h"
 
 #else
 
 # include "isys/dns.h"
-#include "pump/pump.h"
 
 #include "kickstart.h"
 #include "lang.h"
@@ -43,7 +66,7 @@ typedef int int32;
 #ifdef __STANDALONE__
 static FILE * logfile = NULL;
 
-#define FL_TESTING(foo) 0
+#define FL_TESTING(foo) 1
 
 void logMessage(const char * s, ...) {
 	va_list args;
@@ -177,6 +200,23 @@ int nfsGetSetup(char ** hostptr, char ** dirptr) {
 #endif
 
 static void fillInIpInfo(struct networkDeviceConfig * cfg) {
+    int32 * i;
+    char * nm;
+   
+    if (!(cfg->dev.set & PUMP_INTFINFO_HAS_NETMASK)) {
+	i = (int32 *) &cfg->dev.ip;
+
+	if (((*i & 0xFF000000) >> 24) <= 127)
+	    nm = "255.0.0.0";
+	else if (((*i & 0xFF000000) >> 24) <= 191)
+	    nm = "255.255.0.0";
+	else 
+	    nm = "255.255.255.0";
+
+	inet_aton(nm, &cfg->dev.netmask);
+	cfg->dev.set |= PUMP_INTFINFO_HAS_NETMASK;
+    }
+
     if (!(cfg->dev.set & PUMP_INTFINFO_HAS_BROADCAST)) {
 	*((int32 *) &cfg->dev.broadcast) = (*((int32 *) &cfg->dev.ip) & 
 			   *((int32 *) &cfg->dev.netmask)) | 
@@ -199,7 +239,9 @@ void initLoopback(void) {
     strcpy(dev.device, "lo");
     inet_aton("127.0.0.1", &dev.ip);
     inet_aton("255.0.0.0", &dev.netmask);
-    dev.set = PUMP_INTFINFO_HAS_NETMASK | PUMP_INTFINFO_HAS_IP;
+    inet_aton("127.0.0.0", &dev.network);
+    dev.set = PUMP_INTFINFO_HAS_NETMASK | PUMP_INTFINFO_HAS_IP
+		| PUMP_INTFINFO_HAS_NETWORK;
 
     pumpSetupInterface(&dev);
 }
@@ -214,6 +256,37 @@ static void dhcpBoxCallback(newtComponent co, void * ptr) {
     newtEntrySetFlags(c->nsEntry, NEWT_FLAG_DISABLED, NEWT_FLAGS_TOGGLE);
 }
 
+static int getDnsServers(struct networkDeviceConfig * cfg) {
+    int rc;
+    char * ns = "";
+    struct newtWinEntry entry[] = { { N_("Nameserver IP"), &ns, 0 },
+				      { NULL, NULL, 0 } };
+
+    do {
+	rc = newtWinEntries(_("Nameserver"), 
+		_("Your dynamic IP request returned IP configuration "
+		  "information, but it did not include a DNS nameserver. "
+		  "If you know what your nameserver is, please enter it "
+		  "now. If you don't have this information, you can leave "
+		  "this field blank and the install will continue."),
+		40, 5, 10, 25, entry, _("OK"), _("Back"), NULL);
+
+	if (rc == 2) return LOADER_BACK;
+
+	if (ns && *ns && !inet_aton(ns, &cfg->dev.dnsServers[0])) {
+	    newtWinMessage(_("Invalid IP Information"), _("Retry"),
+			_("You entered an invalid IP address."));
+	    rc = 2;
+	} 
+
+    } while (rc == 2);
+
+    cfg->dev.set |= PUMP_NETINFO_HAS_DNS;
+    cfg->dev.numDns = 1;
+
+    return LOADER_OK;
+}
+
 int readNetConfig(char * device, struct networkDeviceConfig * cfg, int flags) {
     newtComponent text, f, okay, back, answer, dhcpCheckbox;
     newtGrid grid, subgrid, buttons;
@@ -345,8 +418,18 @@ int readNetConfig(char * device, struct networkDeviceConfig * cfg, int flags) {
 	    }
 
 	    if (!chptr) {
-		i = 2; 
-		cfg->isDynamic = 1;
+		newCfg.isDynamic = 1;
+#ifdef __STANDALONE__
+		i = 2;
+#else		    
+		if (!(newCfg.dev.set & PUMP_NETINFO_HAS_DNS)) {
+		    logMessage("pump worked, but didn't return a DNS server");
+		    i = getDnsServers(&newCfg);
+		    i = i ? 0 : 2;
+		} else {
+		    i = 2; 
+		}
+#endif		    
 	    } else {
 		logMessage("pump told us: %s", chptr);
 		i = 0;
@@ -354,7 +437,11 @@ int readNetConfig(char * device, struct networkDeviceConfig * cfg, int flags) {
 	}
     } while (i != 2);
 
+#ifdef __STANDALONE__
+    if (!newCfg.isDynamic)
+#endif	  
     cfg->dev = newCfg.dev;
+    cfg->isDynamic = newCfg.isDynamic;
 
     fillInIpInfo(cfg);
 
@@ -380,6 +467,9 @@ int readNetConfig(char * device, struct networkDeviceConfig * cfg, int flags) {
 	writeResolvConf(cfg);
     }
 
+#ifdef __STANDALONE__
+    writeResolvConf(cfg);
+#endif
     return 0;
 }
 
@@ -392,12 +482,29 @@ int configureNetwork(struct networkDeviceConfig * dev) {
     return 0;
 }
 
-int writeNetInfo(const char * fn, struct networkDeviceConfig * dev) {
+int writeNetInfo(const char * fn, struct networkDeviceConfig * dev,
+		 struct knownDevices * kd) {
     FILE * f;
+#ifndef __STANDALONE__
+    int i;
+#endif
 
+#ifndef __STANDALONE__
+    for (i = 0; i < kd->numKnown; i++)
+	if (!strcmp(kd->known[i].name, dev->dev.device)) break;
+#endif
+    
     if (!(f = fopen(fn, "w"))) return -1;
 
     fprintf(f, "DEVICE=%s\n", dev->dev.device);
+
+#ifndef __STANDALONE__
+    if (i < kd->numKnown && kd->known[i].code == CODE_PCMCIA)
+	fprintf(f, "ONBOOT=no\n");
+    else
+#endif
+	fprintf(f, "ONBOOT=yes\n");
+
     if (dev->isDynamic) {
 	fprintf(f, "BOOTPROTO=dhcp\n");
     } else {
@@ -407,6 +514,7 @@ int writeNetInfo(const char * fn, struct networkDeviceConfig * dev) {
 	if (dev->dev.set & PUMP_NETINFO_HAS_GATEWAY)
 	    fprintf(f, "GATEWAY=%s\n", inet_ntoa(dev->dev.gateway));
     }
+
     if (dev->dev.set & PUMP_NETINFO_HAS_HOSTNAME)
 	fprintf(f, "HOSTNAME=%s\n", dev->dev.hostname);
     if (dev->dev.set & PUMP_NETINFO_HAS_DOMAIN)
@@ -459,7 +567,7 @@ int findHostAndDomain(struct networkDeviceConfig * dev, int flags) {
 		  _("Determining host name and domain..."));
 #ifdef __STANDALONE__
 	he = gethostbyaddr( (char *) &dev->dev.ip, sizeof (dev->dev.ip), AF_INET);
-	name = he->h_name;
+	name = he ? he->h_name : 0;
 #else
 	name = mygethostbyaddr(inet_ntoa(dev->dev.ip));
 #endif
@@ -491,20 +599,26 @@ int findHostAndDomain(struct networkDeviceConfig * dev, int flags) {
 }
 
 #ifndef __STANDALONE__
-int kickstartNetwork(char * device, struct networkDeviceConfig * netDev, 
+int kickstartNetwork(char ** devicePtr, struct networkDeviceConfig * netDev, 
 		     char * bootProto, int flags) {
     char ** ksArgv;
     int ksArgc;
     int netSet, rc;
     char * arg, * chptr;
+    char * kshostname=NULL;
     poptContext optCon;
     struct in_addr * parseAddress;
+    int noDns = 0;
+    char * device;
     struct poptOption ksOptions[] = {
 	    { "bootproto", '\0', POPT_ARG_STRING, &bootProto, 0 },
+	    { "device", '\0', POPT_ARG_STRING, devicePtr, 0 },
 	    { "gateway", '\0', POPT_ARG_STRING, NULL, 'g' },
 	    { "ip", '\0', POPT_ARG_STRING, NULL, 'i' },
 	    { "nameserver", '\0', POPT_ARG_STRING, NULL, 'n' },
 	    { "netmask", '\0', POPT_ARG_STRING, NULL, 'm' },
+	    { "nodns", '\0', POPT_ARG_NONE, &noDns, 0 },
+	    { "hostname", '\0', POPT_ARG_STRING, NULL, 'h'},
 	    { 0, 0, 0, 0, 0 }
     };
 
@@ -516,12 +630,12 @@ int kickstartNetwork(char * device, struct networkDeviceConfig * netDev,
 	    ksArgc = 1;
 	}
 
-	optCon = poptGetContext(NULL, ksArgc, ksArgv, ksOptions, 0);
+	optCon = poptGetContext(NULL, ksArgc, (const char **) ksArgv, ksOptions, 0);
 	while ((rc = poptGetNextOpt(optCon)) >= 0) {
 	    parseAddress = NULL;
 	    netSet = 0;
 
-	    arg = poptGetOptArg(optCon);
+	    arg = (char *) poptGetOptArg(optCon);
 
 	    switch (rc) {
 	      case 'g':
@@ -543,9 +657,16 @@ int kickstartNetwork(char * device, struct networkDeviceConfig * netDev,
 		parseAddress = &netDev->dev.netmask;
 		netSet = PUMP_INTFINFO_HAS_NETMASK;
 		break;
+
+	      case 'h':
+		if (kshostname)
+		    free(kshostname);
+		kshostname =  strdup(arg);
+		logMessage("netDev->dev.hostname = %s", kshostname);
+		break;
 	    }
 
-	    if (!inet_aton(arg, parseAddress)) {
+	    if (parseAddress && !inet_aton(arg, parseAddress)) {
 		logMessage("bad ip number in network command: %s", arg);
 		return -1;
 	    }
@@ -563,6 +684,8 @@ int kickstartNetwork(char * device, struct networkDeviceConfig * netDev,
 	}
     }
 
+    device = *devicePtr;
+
     if (!bootProto)
 	bootProto = "dhcp";
 
@@ -579,13 +702,32 @@ int kickstartNetwork(char * device, struct networkDeviceConfig * netDev,
 	    return -1;
 	}
 	netDev->isDynamic = 1;
-    } else {
+    } else if (!strcmp(bootProto, "static")) {
        strcpy(netDev->dev.device, device);
+    } else if (!strcmp(bootProto, "query")) {
+	strcpy(netDev->dev.device, device);
+	readNetConfig("eth0", netDev, flags);
+    } else {
+	newtWinMessage(_("kickstart"), _("OK"),
+		    _("Bad bootproto %s specified in network command"),
+		    bootProto);
+	return -1;
     }
 
     fillInIpInfo(netDev);
     configureNetwork(netDev);
-    findHostAndDomain(netDev, flags);
+
+    logMessage("nodns is %d", noDns);
+
+    if (kshostname) {
+	logMessage("setting ks specified hostname of %s", kshostname);
+	netDev->dev.hostname=strdup(kshostname);
+	netDev->dev.set |= PUMP_NETINFO_HAS_HOSTNAME;
+    }
+
+    if (!noDns)
+	findHostAndDomain(netDev, flags);
+
     writeResolvConf(netDev);
     
     return 0;
@@ -593,14 +735,14 @@ int kickstartNetwork(char * device, struct networkDeviceConfig * netDev,
 #endif
 
 #ifdef __STANDALONE__
-int main(int argc, char **argv) {
+int main(int argc, const char **argv) {
     int netSet, rc;
     int x;
     char * bootProto = NULL;
     char * device = NULL;
     char * hostname = NULL;
     char * domain = NULL;
-    char * arg;
+    const char * arg;
     char path[256];
     char roottext[80];
     poptContext optCon;
@@ -693,7 +835,7 @@ int main(int argc, char **argv) {
 		    netDev->dev.set |= PUMP_NETINFO_HAS_DOMAIN;
 	    }
 	    snprintf(path,256,"/etc/sysconfig/network-scripts/ifcfg-%s",device);
-	    writeNetInfo(path,netDev);
+	    writeNetInfo(path,netDev, NULL);
     } else {
 	    newtInit();
 	    newtCls();
@@ -707,9 +849,11 @@ int main(int argc, char **argv) {
 		    exit(0);
 	    }
 	    if (!device) device="eth0";
-	    readNetConfig(device,netDev,0);
-	    snprintf(path,256,"/etc/sysconfig/network-scripts/ifcfg-%s",device);
-	    writeNetInfo(path,netDev);
+	    strncpy(netDev->dev.device,device,10);
+	    if (readNetConfig(device,netDev,0) != LOADER_BACK) {
+		    snprintf(path,256,"/etc/sysconfig/network-scripts/ifcfg-%s",device);
+		    writeNetInfo(path,netDev, NULL);
+	    }
 	    newtFinished();
     }
     exit(0);
