@@ -1,25 +1,24 @@
 /*
- * Copyright 1999 Red Hat Software, Inc.
+ * Copyright 1999-2001 Red Hat, Inc.
  * 
- *  This program is free software; you can redistribute it and/or modify
- *  it under the terms of the GNU General Public License as published by
- *  the Free Software Foundation; either version 2 of the License, or
- *  (at your option) any later version.
- *
- *  This program is distributed in the hope that it will be useful,
- *  but WITHOUT ANY WARRANTY; without even the implied warranty of
- *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- *  GNU General Public License for more details.
- *
- *  You should have received a copy of the GNU General Public License
- *  along with this program; if not, write to the Free Software
- *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * All Rights Reserved.
+ * 
+ * The above copyright notice and this permission notice shall be included in
+ * all copies or substantial portions of the Software.
+ * 
+ * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
+ * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
+ * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
+ * OPEN GROUP BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
+ * AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
+ * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
+ * 
+ * Except as contained in this notice, the name of Red Hat shall not be
+ * used in advertising or otherwise to promote the sale, use or other dealings
+ * in this Software without prior written authorization from Red Hat.
  *
  */
 
-/* This probably doesn't work on anything other then Ethernet! It may work
-   on PLIP as well, but ARCnet and Token Ring are unlikely at best. */
-
 #include <arpa/inet.h>
 #include <ctype.h>
 #include <errno.h>
@@ -47,6 +46,7 @@
 #include <sys/utsname.h>
 #include <sys/wait.h>
 #include <syslog.h>
+#include <time.h>
 #include <unistd.h>
 
 #include "config.h"
@@ -85,11 +85,12 @@ struct command {
 	    char hostname[1024];
 	    char domain[1024];
 	    char bootFile[1024];
+	    char nisDomain[1024];
 	} status;
     } u;
 };
 
-static int openControlSocket(char * configFile);
+static int openControlSocket(char * configFile, struct pumpOverrideInfo * override);
 
 char * readSearchPath(void) {
     int fd;
@@ -124,6 +125,8 @@ char * readSearchPath(void) {
 	    *buf = '\0';
 	    return strdup(start);
 	}
+
+	while (*start && (*start != '\n')) start++;
     }
 
     return NULL;
@@ -158,10 +161,31 @@ static void createResolvConf(struct pumpNetIntf * intf, char * domain,
 
     if (domain && isSearchPath) {
 	fprintf(f, "search %s\n", domain);
+    } else if (domain && !strchr(domain, '.')) {
+	fprintf(f, "search %s\n", domain);
     } else if (domain) {
 	fprintf(f, "search");
 	chptr = domain;
 	do {
+	    /* If there is a single . in the search path, write it out
+	     * only if the toplevel domain is com, edu, gov, mil, org,
+	     * net 
+	     */
+	    /* Don't do that! It breaks virtually all installations
+	     * in Europe.
+	     * Besides, what's wrong with some company assigning hostnames
+	     * in the ".internal" TLD?
+	     * What exactly was this supposed to accomplish?
+	     * Commented out --bero
+	     */
+/*	    if (!strchr(strchr(chptr, '.') + 1, '.')) {
+		char * tail = strchr(chptr, '.');
+		if (strcmp(tail, ".com") && strcmp(tail, ".edu") &&
+		    strcmp(tail, ".gov") && strcmp(tail, ".mil") &&
+		    strcmp(tail, ".net") && 
+		    strcmp(tail, ".org") && strcmp(tail, ".int")) break;
+	    } */
+
 	    fprintf(f, " %s", chptr);
 	    chptr = strchr(chptr, '.');
 	    if (chptr) {
@@ -183,6 +207,39 @@ static void createResolvConf(struct pumpNetIntf * intf, char * domain,
     endhostent();
 }
 
+void setupDomain(struct pumpNetIntf * intf, 
+		 struct pumpOverrideInfo * override) {
+    int bufSize = 128;
+    char * buf = NULL;
+
+    if (override->flags & OVERRIDE_FLAG_NONISDOMAIN)
+	return;
+    if (!(intf->set & PUMP_NETINFO_HAS_NISDOMAIN))
+	return;
+
+    buf = malloc(bufSize);
+    while (getdomainname(buf, bufSize)) {
+	if (errno != EINVAL) {
+	    syslog(LOG_ERR, "failed to get domainname: %s", strerror(errno));
+	    return;
+	}
+
+	buf += 128;
+    }
+
+    /* if the domainname is set, then don't override it */
+    if (strcmp(buf, "localdomain") && strcmp(buf, "")) {
+	return;
+    }
+
+    if (setdomainname(intf->domain, strlen(intf->domain))) {
+	syslog(LOG_ERR, "failed to set domainname: %s", strerror(errno));
+	return ;
+    }
+
+    return;
+}
+
 void setupDns(struct pumpNetIntf * intf, struct pumpOverrideInfo * override) {
     char * hn, * dn = NULL;
     struct hostent * he;
@@ -225,10 +282,85 @@ void setupDns(struct pumpNetIntf * intf, struct pumpOverrideInfo * override) {
     }
 }
 
-static void runDaemon(int sock, char * configFile) {
+static void callIfupPost(struct pumpNetIntf* intf) {
+    pid_t child;
+    char * argv[3];
+    char arg[64];
+
+    argv[0] = "/etc/sysconfig/network-scripts/ifup-post";
+    snprintf(arg,64,"ifcfg-%s",intf->device);
+    argv[1] = arg;
+    argv[2] = NULL;
+
+    if (!(child = fork())) {
+	/* send the script to init */
+	if (fork()) _exit(0);
+
+	execvp(argv[0], argv);
+
+	syslog(LOG_ERR,"failed to run %s: %s", argv[0], strerror(errno));
+
+	_exit(0);
+    }
+
+    waitpid(child, NULL, 0);
+}
+
+static void callScript(char* script,int msg,struct pumpNetIntf* intf) {
+    pid_t child;
+    char * argv[20];
+    char ** nextArg;
+    char * class, * chptr;
+
+    if (!script) return;
+
+    argv[0] = script;
+    argv[2] = intf->device;
+    nextArg = argv + 3;
+
+    switch (msg) {
+	case PUMP_SCRIPT_NEWLEASE:
+	    class = "up";
+	    chptr = inet_ntoa(intf->ip);
+	    *nextArg = alloca(strlen(chptr) + 1);
+	    strcpy(*nextArg, chptr);
+	    nextArg++;
+	    break;
+
+	case PUMP_SCRIPT_RENEWAL:
+	    class = "renewal";
+	    chptr = inet_ntoa(intf->ip);
+	    *nextArg = alloca(strlen(chptr) + 1);
+	    strcpy(*nextArg, chptr);
+	    nextArg++;
+	    break;
+
+	case PUMP_SCRIPT_DOWN:
+	    class = "down";
+	    break;
+    }
+
+    argv[1] = class;
+    *nextArg = NULL;
+
+    if (!(child = fork())) {
+	/* send the script to init */
+	if (fork()) _exit(0);
+
+	execvp(argv[0], argv);
+
+	syslog(LOG_ERR,"failed to run %s: %s", argv[0], strerror(errno));
+
+	_exit(0);
+    }
+
+    waitpid(child, NULL, 0);
+}
+
+static void runDaemon(int sock, char * configFile, struct pumpOverrideInfo * overrides) {
     int conn;
     struct sockaddr_un addr;
-    int addrLength;
+    int addrLength = sizeof(struct sockaddr_un);
     struct command cmd;
     struct pumpNetIntf intf[20];
     int numInterfaces = 0;
@@ -236,10 +368,11 @@ static void runDaemon(int sock, char * configFile) {
     int closest;
     struct timeval tv;
     fd_set fds;
-    struct pumpOverrideInfo * overrides = NULL;
     struct pumpOverrideInfo emptyOverride, * o;
 
-    readPumpConfig(configFile, &overrides);
+    if (!overrides)
+        readPumpConfig(configFile, &overrides);
+
     if (!overrides) {
 	overrides = &emptyOverride;
 	overrides->intf.device[0] = '\0';
@@ -258,14 +391,14 @@ static void runDaemon(int sock, char * configFile) {
 			       (intf[closest].renewAt > intf[i].renewAt)))
 		    closest = i;
 	    if (closest != -1) {
-		tv.tv_sec = intf[closest].renewAt - time(NULL);
+		tv.tv_sec = intf[closest].renewAt - pumpUptime();
 		if (tv.tv_sec <= 0) {
 		    if (pumpDhcpRenew(intf + closest)) {
 			syslog(LOG_INFO,
 				"failed to renew lease for device %s",
 				intf[closest].device);
 
-			if ((intf[closest].renewAt += time(NULL) + 30) >
+			if ((intf[closest].renewAt += pumpUptime() + 30) >
 			    intf[closest].leaseExpiration) {
 			    o = overrides;
 			    while (*o->intf.device &&
@@ -275,12 +408,16 @@ static void runDaemon(int sock, char * configFile) {
 			    if (!*o->intf.device) o = overrides;
 
 			    intf[closest].set &= ~PUMP_INTFINFO_HAS_LEASE;
-			    if (pumpDhcpRun(intf[closest].device, 0, 
+
+			    if (pumpDhcpRun(intf[closest].device, 
+				  intf[closest].flags, 
 				  intf[closest].reqLease,
 				  intf[closest].set & PUMP_NETINFO_HAS_HOSTNAME
 				    ? intf[closest].hostname : NULL,
 				  intf + closest, o)) {
 				if (numInterfaces == 1) {
+				    callScript(o->script, PUMP_SCRIPT_DOWN,
+					       &intf[closest]);
 				    syslog(LOG_INFO,
 					    "terminating as there are no "
 					    "more devices under management");
@@ -289,9 +426,15 @@ static void runDaemon(int sock, char * configFile) {
 
 				intf[i] = intf[numInterfaces - 1];
 				numInterfaces--;
-			    }
+			    } else {
+				callScript(o->script, PUMP_SCRIPT_NEWLEASE,
+					   &intf[closest]);
+                            }
 			}
-
+		    } else {
+			callScript(o->script, PUMP_SCRIPT_RENEWAL,
+				   &intf[closest]);
+			callIfupPost(&intf[closest]);
 		    }
 
 		    continue;	    /* recheck timeouts */
@@ -310,8 +453,10 @@ static void runDaemon(int sock, char * configFile) {
 
 	    switch (cmd.type) {
 	      case CMD_DIE:
-		for (i = 0; i < numInterfaces; i++)
+		for (i = 0; i < numInterfaces; i++) {
 		    pumpDhcpRelease(intf + i);
+		    callScript(o->script, PUMP_SCRIPT_DOWN, &intf[i]);
+		}
 
 		syslog(LOG_INFO, "terminating at root's request");
 
@@ -335,15 +480,20 @@ static void runDaemon(int sock, char * configFile) {
 			        intf + numInterfaces, o)) {
 		    cmd.u.result = 1;
 		} else {
-		    pumpSetupInterface(intf);
+		    pumpSetupInterface(intf + numInterfaces);
+		    i = numInterfaces;
 
-		    syslog(LOG_INFO, "configured interface %s", intf->device);
+		    syslog(LOG_INFO, "configured interface %s", intf[i].device);
 
-		    if (intf->set & PUMP_NETINFO_HAS_GATEWAY) {
-			pumpSetupDefaultGateway(&intf->gateway);
-		    }
+		    if ((intf[i].set & PUMP_NETINFO_HAS_GATEWAY) &&
+			 !(o->flags & OVERRIDE_FLAG_NOGATEWAY))
+			pumpSetupDefaultGateway(&intf[i].gateway);
 
-		    setupDns(intf, o);
+		    setupDns(intf + i, o);
+		    setupDomain(intf + i, o);
+
+		    callScript(o->script, PUMP_SCRIPT_NEWLEASE, 
+			       intf + numInterfaces);
 
 		    cmd.u.result = 0;
 		    numInterfaces++;
@@ -355,8 +505,13 @@ static void runDaemon(int sock, char * configFile) {
 		    if (!strcmp(intf[i].device, cmd.u.renew.device)) break;
 		if (i == numInterfaces)
 		    cmd.u.result = RESULT_UNKNOWNIFACE;
-		else
+		else {
 		    cmd.u.result = pumpDhcpRenew(intf + i);
+		    if (!cmd.u.result) {
+			callScript(o->script, PUMP_SCRIPT_RENEWAL, intf + i);
+			callIfupPost(intf + i);
+		    }
+		}
 		break;
 
 	      case CMD_STOPIFACE:
@@ -366,6 +521,7 @@ static void runDaemon(int sock, char * configFile) {
 		    cmd.u.result = RESULT_UNKNOWNIFACE;
 		else {
 		    cmd.u.result = pumpDhcpRelease(intf + i);
+		    callScript(o->script, PUMP_SCRIPT_DOWN, intf + i);
 		    if (numInterfaces == 1) {
 			cmd.type = CMD_RESULT;
 			write(conn, &cmd, sizeof(cmd));
@@ -404,6 +560,11 @@ static void runDaemon(int sock, char * configFile) {
 			    intf->bootFile, sizeof(cmd.u.status.bootFile));
 		    cmd.u.status.bootFile[sizeof(cmd.u.status.bootFile) - 1] = 
 		    							'\0';
+		    if (intf[i].set & PUMP_NETINFO_HAS_NISDOMAIN)
+			strncpy(cmd.u.status.nisDomain,
+			    intf->nisDomain, sizeof(cmd.u.status.nisDomain));
+		    cmd.u.status.nisDomain[sizeof(cmd.u.status.nisDomain)-1] = 
+		    							'\0';
 		}
 
 	      case CMD_STATUS:
@@ -422,7 +583,7 @@ static void runDaemon(int sock, char * configFile) {
     exit(0);
 }
 
-static int openControlSocket(char * configFile) {
+static int openControlSocket(char * configFile, struct pumpOverrideInfo * override) {
     struct sockaddr_un addr;
     int sock;
     size_t addrLength;
@@ -473,13 +634,23 @@ static int openControlSocket(char * configFile) {
 
 	openlog("pumpd", LOG_PID, LOG_DAEMON);
 	{
-	    time_t t;
+	    time_t now,upt;
+	    int updays,uphours,upmins,upsecs;
 
-	    t = time(NULL);
-	    syslog(LOG_INFO, "starting at %s\n", ctime(&t));
+	    now = time(NULL);
+	    upt = pumpUptime();
+	    if (now <= upt)
+		syslog(LOG_INFO, "starting at %s\n", ctime(&now));
+	    else {
+		upsecs = upt % 60;
+		upmins = (upt / 60) % 60;
+		uphours = (upt / 3600) % 24;
+		updays = upt / 86400;
+		syslog(LOG_INFO, "starting at (uptime %d days, %d:%02d:%02d) %s\n", updays, uphours, upmins, upsecs, ctime(&now));
+	    }
 	}
 
-	runDaemon(sock, configFile);
+	runDaemon(sock, configFile, override);
     }
 
     waitpid(child, &status, 0);
@@ -496,8 +667,9 @@ static int openControlSocket(char * configFile) {
 }
 
 void printStatus(struct pumpNetIntf i, char * hostname, char * domain,
-		 char * bootFile) {
+		 char * bootFile, char * nisDomain) {
     int j;
+    time_t now,upnow,localAt,localExpiration;
 
     printf("Device %s\n", i.device);
     printf("\tIP: %s\n", inet_ntoa(i.ip));
@@ -526,6 +698,9 @@ void printStatus(struct pumpNetIntf i, char * hostname, char * domain,
 	printf("\n");
     }
 
+    if (i.set & PUMP_NETINFO_HAS_NISDOMAIN)
+	printf("\tNIS Domain: %s\n", nisDomain);
+
     if (i.numLog) {
 	printf("\tLogservers:");
 	for (j = 0; j < i.numLog; j++)
@@ -562,12 +737,17 @@ void printStatus(struct pumpNetIntf i, char * hostname, char * domain,
     }
 
     if (i.set & PUMP_INTFINFO_HAS_LEASE) {
-	printf("\tRenewal time: %s", ctime(&i.renewAt)); 
-	printf("\tExpiration time: %s", ctime(&i.leaseExpiration)); 
+	upnow = pumpUptime();
+	tzset();
+	now = time(NULL);
+	localAt = now + (i.renewAt - upnow);
+	localExpiration = now + (i.leaseExpiration - upnow);
+	printf("\tRenewal time: %s", ctime(&localAt)); 
+	printf("\tExpiration time: %s", ctime(&localExpiration)); 
     }
 }
 
-int main (int argc, char ** argv) {
+int main (int argc, const char ** argv) {
     char * device = "eth0";
     char * hostname = "";
     poptContext optCon;
@@ -576,7 +756,9 @@ int main (int argc, char ** argv) {
     int flags = 0;
     int lease = 12;
     int killDaemon = 0;
-    int release = 0, renew = 0, status = 0, lookupHostname = 0;
+    int winId = 0;
+    int release = 0, renew = 0, status = 0, lookupHostname = 0, nodns = 0;
+    int nogateway = 0;
     struct command cmd, response;
     char * configFile = "/etc/pump.conf";
     struct pumpOverrideInfo * overrides;
@@ -602,6 +784,12 @@ int main (int argc, char ** argv) {
 			N_("Force immediate lease renewal"), NULL },
 	    { "status", 's', POPT_ARG_NONE, &status, 0,
 			N_("Display interface status"), NULL },
+	    { "no-dns", 'd', POPT_ARG_NONE, &nodns, 0,
+			N_("Don't update resolv.conf"), NULL },
+	    { "no-gateway", '\0', POPT_ARG_NONE, &nogateway, 0,
+			N_("Don't set a gateway for this interface"), NULL },
+	    { "win-client-ident", '\0', POPT_ARG_NONE, &winId, 0,
+			N_("Set the client identifier to match Window's") },
 	    /*{ "test", 't', POPT_ARG_NONE, &test, 0,
 			N_("Don't change the interface configuration or "
 			   "run as a deamon.") },*/
@@ -609,6 +797,9 @@ int main (int argc, char ** argv) {
 	    { NULL, '\0', 0, NULL, 0 }
         };
 
+    memset(&cmd, 0, sizeof(cmd));
+    memset(&response, 0, sizeof(response));
+
     optCon = poptGetContext(PROGNAME, argc, argv, options,0);
     poptReadDefaultConfig(optCon, 1);
 
@@ -634,10 +825,16 @@ int main (int argc, char ** argv) {
 
     if (test)
 	flags = PUMP_FLAG_NODAEMON | PUMP_FLAG_NOCONFIG;
+    if (winId)
+	flags |= PUMP_FLAG_WINCLIENTID;
     if (lookupHostname)
 	flags |= PUMP_FLAG_FORCEHNLOOKUP;
+    if (nodns)
+	overrides->flags |= OVERRIDE_FLAG_NODNS;
+    if (nogateway)
+	overrides->flags |= OVERRIDE_FLAG_NOGATEWAY;
 
-    cont = openControlSocket(configFile);
+    cont = openControlSocket(configFile, overrides);
     if (cont < 0) 
 	exit(1);
 
@@ -665,7 +862,7 @@ int main (int argc, char ** argv) {
 
     if (response.type == CMD_RESULT && response.u.result &&
 	    cmd.type == CMD_STARTIFACE) {
-	cont = openControlSocket(configFile);
+	cont = openControlSocket(configFile, overrides);
 	if (cont < 0) 
 	    exit(1);
 	write(cont, &cmd, sizeof(cmd));
@@ -679,7 +876,8 @@ int main (int argc, char ** argv) {
 	}
     } else if (response.type == CMD_STATUS) {
 	printStatus(response.u.status.intf, response.u.status.hostname, 
-		    response.u.status.domain, response.u.status.bootFile);
+		    response.u.status.domain, response.u.status.bootFile,
+		    response.u.status.nisDomain);
     }
 
     return 0;
